/*
2010 KCNT roboken A team
Recieve data from RF module
PIC16F88

ver 1.0.0

RF Module to UART
(all channel)

final update:10/09/21
*/

#define LED1 PORTB.F3
#define LED2 PORTB.F4

#define IN1 PORTB.F6
#define IN2 PORTB.F7



unsigned short flag;
//flag bit parameter
/*
0bit: RF falling edge
1bit: Neutral

3bit: Neutral transmit 0x00

7bit: error flag
*/

unsigned short ch;

unsigned short tmp;

unsigned short count;

unsigned short data;

unsigned short re_count;
unsigned short TX_past;

unsigned short parity;

unsigned short time_H;
unsigned short time_L;

void interrupt(){
  if(INTCON.RBIF){
    if(TMR1H>=3){//18ms stop
      ch=0;
      flag&=0x80;
    }
    if(IN1==1){
      if(ch==0){//first rising edge
        TMR1L=0;
        TMR1H=0;
        T1CON.TMR1ON=1;
        //Timer counter reset and start
      }else if(ch==1){//3rd rising edge
        T1CON.TMR1ON=0;//counter stop
        time_H=TMR1H;//save time
        time_L=TMR1L;
        TMR1L=0;//reset count
        TMR1H=0;
        T1CON.TMR1ON=1;//count start
        flag|=0x01;//recieve flag
        ch++;
      }
    }else if(IN2==1){
      if((ch==0)||(ch==2)){//2nd ,4th rising edge
        T1CON.TMR1ON=0;//stop count
        time_H=TMR1H;//save time
        time_L=TMR1L;
        TMR1L=0;//reset count
        TMR1H=0;
        T1CON.TMR1ON=1;//count start
        flag|=0x01;//recieve flag
        ch++;
      }
    }else{
      if(ch==3){//final falling edge
        T1CON.TMR1ON=0;//counter stop
        time_H=TMR1H;//save count
        time_L=TMR1L;
        T1CON.TMR1ON=1;//count start
        flag|=0x01;//recieve flag
        ch++;
      }else if(ch<=2){//error edge
        T1CON.TMR1ON=0;//stop count
        time_H=TMR1H;//save time
        time_L=TMR1L;
        TMR1L=0;//reset count
        TMR1H=0;
        T1CON.TMR1ON=1;//count start
        flag|=0x01;//recieve flag
        ch++;
      }
    }
  }
  INTCON.RBIF=0;//flag reset
  return;
}

unsigned short parity_check(unsigned short x){
  while(x&0xfe){
    if(x&0x01){
      if(x&0x02){
        x=x>>2;
      }else{
        x=x>>1;
        x+=1;
      }
    }else{
      x=x>>1;
    }
  }
  return x;
}


void init(){
  //8MHz
  OSCCON=0x70;

  //I/O config
  OPTION_REG=0x51;
  TRISB=0xc4;//0b11000100
  INTCON=0xc8;
  ANSEL=0x00;

  //TIMER1 config
  T1CON=0x31;

  //UART config
  TXSTA=0x24;//00100100
  RCSTA=0x90;
  SPBRG=0x33;

  //variable value reset
  data=0;
  count=0;
  tmp=0;
  TX_past=0;
  parity=0;
  return;
}

main(){
  init();
  while(1){
    if(flag&0x01){//data recieve flag
      if(time_H==1){//recieve time from 1ms to 2ms
        flag&=0xfe;
        tmp=time_L;
        count=0;
        if(tmp>31){//from time to value
          tmp-=31;
          count++;
          while(tmp>25){
            tmp-=25;
            count++;
            if(count>=8){
              break;
            }
          }
        }
        if(count<=3){
        }else if(count==4){//Neutral
          flag|=0x02;
        }else{
          count-=1;
        }
        switch(ch){
          case 1://ch1
          case 2://ch2
            data=data<<3;
            data+=count;
            break;
          case 3://CH3
            data=data<<2;
            if(count&0x01){//parity
              parity=1;
            }else{
              parity=0;
            }
            count=count>>1;
            data+=count;
            break;
          case 4://CH4 (last pulse)
            if(flag&0x02){//Neutral
              flag&=0xfd;//Neutral flag reset
              if(TX_past){//transmit 0x00 (once time)
                TXREG=0x00;
                TX_past=0x00;
              }
              LED1=0;
            }else{
              if(parity_check(data)==parity){//parity_check OK
                LED1=1;
                if(TX_past!=data){//Transmit data
                    TX_past=data;
                    TXREG=data;
                    re_count=count;
                }else if(re_count!=count){//New recieve UART check
                  TXREG=data;
                  re_count=count;
                }
                flag=0;
//              }else{//parity error
//                flag|=0x80;
//                TXREG=0x00;
              }
            }
            data=0;
            ch=0;
            break;
          default://CH over
            break;
        }
      }else{//over 1.9ms
        if(TX_past){//transmit 0x00 (once time)
          TXREG=0x00;
          TX_past=0x00;
          LED1=0;
        }
        flag&=0xfe;
      }
    }
  }
}
